#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <limits.h>
#include <glob.h>
#include <pwd.h>
#include <grp.h>
#include "dftpd.h"

/* libdftp dependencies */

int dftp_ugetbyid(int, unsigned int, struct dftp_uent *);
int dftp_ggetbyid(int, unsigned int, struct dftp_gent *);
void dftp_ufree(struct dftp_uent *);
void dftp_gfree(struct dftp_gent *);

/* dftpd.c dependencies */

int open_dataconn(char, char, char *);
void close_dataconn(int);
void ftplog(const char *, const char *, ...);

extern struct dftp_vars var;
extern int sockfd, passwdfile, groupfile;

/* aux.c dependencies */

char **parsestr(char *);
void _status(unsigned short);
int checklog();

/* srv-common.c dependencies */

void writes(const char *,...);

void printperms(mode_t mode)
{
   if(S_ISLNK(mode)) writes("l");
   else if(S_ISDIR(mode)) writes("d");
   else if(S_ISBLK(mode)) writes("b");
   else if(S_ISCHR(mode)) writes("c");
   else if(S_ISSOCK(mode)) writes("s");
   else if(S_ISFIFO(mode)) writes("f");
   else writes("-");
   if(mode&S_IRUSR) writes("r"); else writes("-");
   if(mode&S_IWUSR) writes("w"); else writes("-");
   if(mode&S_ISUID) writes("s");
   else if(mode&S_IXUSR) writes("x"); else writes("-");
   if(mode&S_IRGRP) writes("r"); else writes("-");
   if(mode&S_IWGRP) writes("w"); else writes("-");
   if(mode&S_ISGID) writes("s");
   else if(mode&S_IXGRP) writes("x"); else writes("-");
   if(mode&S_IROTH) writes("r"); else writes("-");
   if(mode&S_IWOTH) writes("w"); else writes("-");
   if(mode&S_ISVTX) writes("t");
   else if(mode&S_IXOTH) writes("x"); else writes("-");
}

char *resolveuid(unsigned int uid)
{
   int result; struct dftp_uent dentry; struct passwd *entry;
   static char username[30]; static unsigned int cacheuid = 0;

   if(!uid)
   {
      sprintf(username,"root"); return username;
   }
   if(cacheuid == uid) return username;
   memset(username,0,30); cacheuid = uid;
   if(var.userclass != REAL)
   {
      flock(passwdfile,LOCK_EX);
      result = dftp_ugetbyid(passwdfile, uid, &dentry);
      flock(passwdfile,LOCK_UN);
      if(result)
      {
         sprintf(username,"%-8s",dentry.name); dftp_ufree(&dentry);
         return username;
      }
   }
   else if((entry = getpwuid(uid)))
   {
      sprintf(username,"%-8s",entry->pw_name); return username;
   }
   sprintf(username,"%u",uid); return username;
}

char *resolvegid(unsigned int gid)
{
   int result; struct dftp_gent dentry; struct group *entry;
   static char groupname[30]; static unsigned int cachegid = 0;

   if(!gid)
   {
      sprintf(groupname,"root"); return groupname;
   }
   if(cachegid == gid) return groupname;
   memset(groupname,0,30); cachegid = gid;
   if(var.userclass != REAL)
   {
      flock(groupfile,LOCK_EX);
      result = dftp_ggetbyid(groupfile, gid, &dentry);
      flock(groupfile,LOCK_UN);
      if(result)
      {
         sprintf(groupname,"%-8s",dentry.name); dftp_gfree(&dentry);
         return groupname;
      }
   }
   else if((entry = getgrgid(gid)))
   {
      sprintf(groupname,"%-8s",entry->gr_name); return groupname;
   }
   sprintf(groupname,"%u",gid); return groupname;
}

char **getlsopts(char *args, int *opts)
{
   char **argv, **targets = NULL; int i, arg, targetcount = 0;
   struct stat statbuf;

   if(!(argv = parsestr(args)))
   {
      targets = realloc(targets,2*sizeof(char *));
      targets[0] = malloc(4); strcpy(targets[0],"./*");
      targets[1] = NULL; return targets;
   }
   for(arg = 0;argv[arg];arg++)
   {
      if(*argv[arg] == '-')
         for(i=1;i<strlen(argv[arg]);i++)
            switch(argv[arg][i])
            {
               case 'l': *opts |= LS_LONG; break;
               case '1': if(*opts&LS_LONG) *opts ^= LS_LONG; break;
               case 'n': *opts |= LS_NORESOLVE; break;
               case 'A': *opts |= LS_ALMOSTALL; break;
               case 'r': *opts |= LS_REVERSE; break;
               case 'R': *opts |= LS_RECURSIVE; break;
            }
      else
      {
         targets = realloc(targets,++targetcount*sizeof(char *));
         targets[targetcount-1] = malloc(strlen(argv[arg])+1);
         strcpy(targets[targetcount-1],argv[arg]);
         if(!lstat(argv[arg],&statbuf) && S_ISDIR(statbuf.st_mode))
         {
            targets[targetcount-1] = realloc(targets[targetcount-1],strlen(targets[targetcount-1])+3);
            strcat(targets[targetcount-1],"/*");
         }
      }
   }
   for(i=0;argv[i];i++) free(argv[i]); free(argv);
   if(!targetcount)
   {
      targets = malloc(++targetcount*sizeof(char *));
      targets[0] = malloc(4); strcpy(targets[0],"./*");
   }
   targets = realloc(targets,++targetcount*sizeof(char *));
   targets[targetcount-1] = NULL; return targets;
}

void listcmd(int type, char *args)
{
   int datasock, tempfd = sockfd, opts = 0, glob_opts = 0, i, tgtnum;
   struct stat statbuf; char buf[PATH_MAX+1], *ptr, **targets; glob_t glt; struct tm *tme;
   int loopstart = 0, loopend, loopinterval = 1;

   if(!checklog()) return;

   if(!type) opts |= LS_LONG;
   targets = getlsopts(args,&opts);
   datasock = open_dataconn('A',ID_CONTROL,"/bin/ls");
   if(datasock < 0) return;
   shutdown(datasock,0);
   sockfd = datasock; memset(buf,0,PATH_MAX+1);

#ifdef GLOB_BRACE
   glob_opts |= GLOB_BRACE;
#endif
   if(!(opts&LS_ALMOSTALL)) glob_opts |= GLOB_PERIOD;
   for(tgtnum = 0;targets[tgtnum];tgtnum++)
   {
      glob(targets[tgtnum],glob_opts,NULL,&glt);
      if(opts&LS_REVERSE)
      {
         loopstart=glt.gl_pathc-1; loopend = -1; loopinterval = -1;
      }
      else loopend = glt.gl_pathc;
      for(i=loopstart;i!=loopend;i += loopinterval)
      {
         if(opts & LS_LONG)
         {
            lstat(glt.gl_pathv[i],&statbuf); printperms(statbuf.st_mode);
            tme = localtime(&statbuf.st_mtime); strftime(buf,15,"%b %e %R",tme);
            if(!(ptr = strrchr(glt.gl_pathv[i],'/'))) ptr = glt.gl_pathv[i]; else ptr++;
            writes(" %3i %-8s %-8s %8u %s %s",statbuf.st_nlink,resolveuid(statbuf.st_uid),resolvegid(statbuf.st_gid),statbuf.st_size,buf,ptr);
         }
         else writes("%s",strrchr(glt.gl_pathv[i],'/')+1);
         if(S_ISLNK(statbuf.st_mode))
         {
            memset(buf,0,PATH_MAX+1); readlink(glt.gl_pathv[i],buf,PATH_MAX);
            writes(" -> %s",buf);
         }
         writes("\r\n");
      }
      globfree(&glt);
   }
   sockfd = tempfd; _status(226);
   close_dataconn(datasock);
   for(i=0;targets[i];i++) free(targets[i]); free(targets);
}

